__all__ = ['benchmark_func', 'experiment', 'hyperheuristic', 'metaheuristic', 'operators', 'population', 'tools',
           'visualisation']
# Under revision: 'characterisation', 'visualisation'
__version__ = "1.1.4"
